#!/usr/bin/python
# Use AIC score as Fitness
# Family Competition
# Do NOT Impose Self Links
# Do not impose the constraint -- Indegree(i) >= 1 (Do not count self-links)
# Use two SAs
# Without Power Law
######################################################################
## Gene Network -- GA and SA
## Family Competition
## Input: MicroArray Data filen
## 
######################################################################
import My_Graph
import My_Sub
import random
import math
from sys import argv
from threading import Thread
import thread
import os
import Queue
import getpass
import time
import string
import regression
import pexpect
import sys
import cPickle

trainingfilen = argv[1]
global programn
#programn = 'SA_TrueModelONEGENE'
#program_slow = 'SA_TrueModelONEGENE'
#program_fast = 'SA_TrueModelONEGENE'

programn = 'a.out'
program_slow = 'a.out'
program_fast = 'a.out'

#ssh_workers = []
#for i in xrange(32, 34, 2):
#    num = str(i)
#    #num = '0' * (3 - len(num)) + num
#    ssh_workers.append('node' + num)

N_FACTOR = 2
MAXLINKS = 3

main_path = os.getcwd()

# 50 for the noise-free simulated data (under AIC)
# 0.1 for the SNR? = 10 one
# 0.05 for the SNR? = 4 one

# 1 for the SSEVAR version
W_SSE = 1

Variance = 0.671    # for the 9-gene network
global Link_Density
global MutationRate
global CrossOverRate
Link_Density=0.2
MutationRate=0.005
CrossOverRate=0.30
P_Parent = 0.5  # proportion of parent chromosomes to keep
MaxInDegree=51  # <51 on simulated data (<52 time points), 18 on true data (18 time points)
class chromosome:
    # number of genes, number of cells, cells, Wij, alpha
    def __init__(self, n_gene, n_cells, cells=None, Wij=None, alpha=None, beta=None, fitness=0, AIC=0, BIC=0, SSE = 0, RSSE = 0, SSEVAR = 0):
        self.n_cells=n_cells
        self.n_gene=n_gene
        self.atleastone = 1
        if not cells:
            # randomly generate a structure
            self.Random_Assignment()
        else:
            self.cells=cells
        self.narcs = self.Arcs() # + self.n_gene - N_FACTOR   # including self links
        self.fitness = fitness  # BIC score or AIC score
        self.AIC = AIC  # AIC score
        self.BIC = BIC  # BIC score
        self.SSE = SSE  # Sum of Square Error
        self.RSSE = RSSE    # Relative Sum of Square Error
        self.SSEVAR = SSEVAR
        if (not Wij) or (not alpha) or (not beta):
            # randomly generate Wij, alpha, beta
            self.Generate_parameters()
        else:
            self.Wij=Wij
            self.alpha=alpha
            self.beta=beta
        self.GetInDegree()

    def Arcs(self):
        arcs=0
        for cell in self.cells:
            if cell:
                arcs+=1
        return arcs

    def Generate_Network(self):
        g = random.Random()
        Net = []
        self.narcs=0
        self.InDeg = [0] * self.n_gene
        for i in xrange(N_FACTOR):  # factors do not have any incoming link
            Net += [0 for j in xrange(MAXLINKS)]
        for i in xrange(N_FACTOR, self.n_gene):    # for each gene, generate the number of links (excluding self link)
            for j in xrange(MAXLINKS): # at most MAXLINKS links (self link not included)
                Net.append(g.randint(0,1))  # 0 or 1
                self.InDeg[i] += Net[-1]                
        return Net
        
    def Random_Assignment(self):
        Net = self.Generate_Network()
        self.cells=Net
#        self.n_cells = len(Net)

    def Flip(self, p):
        if self.cells[p]==1:
            self.cells[p]=0
        else:
            self.cells[p]=1
            
    def Mutate(self):
        g = random.Random()
        for i in range(N_FACTOR*MAXLINKS, self.n_gene*MAXLINKS):
            if g.random() > MutationRate:
                continue
#            if self.cells[i]==0 and self.InDeg[j] >= MaxInDegree:
#                continue
            self.Flip(i)
            if self.cells[i]==0:    # 1-->0
#                self.Wij[i*self.n_gene+j]=0.0
                self.narcs -= 1   # number of arcs - 1
            else:   # 0-->1
#                self.Wij[i*self.n_gene+j] = 0.0 # g.random()-0.5  # randomly assign Wij[p]
                self.narcs += 1   # number of arcs + 1
        self.GetInDegree()

    def Generate_parameters(self):
        g = random.Random()
        self.Wij = []
        self.alpha = []
        self.beta = []
        for i in range(self.n_gene * self.n_gene):
            self.Wij.append(0.0)
        for i in range(self.n_gene):
            self.alpha.append(0.0)
            self.beta.append(0.0)
            
    def GetInDegree(self):
        self.InDeg = [0] * self.n_gene
        for i in xrange(N_FACTOR, self.n_gene):
            for j in xrange(MAXLINKS):
                self.InDeg[i] += self.cells[i*MAXLINKS+j]
                   
    def hist(self):
        self.atleastone = 1
        self.histogram = {}
        for i in xrange(N_FACTOR, self.n_gene):
            if self.InDeg[i]==0:   # ignore gene with indegree == 0
                self.atleastone = 0
                continue
            if not self.histogram.has_key(self.InDeg[i]):
                self.histogram[self.InDeg[i]] = 0
            self.histogram[self.InDeg[i]] += 1
#        for key in self.Lhistogram.keys():
#            self.Lhistogram[key] = math.log(self.Lhistogram[key])

    ## NOTE: no power law, and isolated genes are not allowed.
    def Valid(self):
        indegree_list = self.InDeg[N_FACTOR:N_Gene]
        for i in indegree_list:
            if i == 0:
                return 0

        return 1
        

def Init_Chroms(Chroms, N, n_gene):
    for i in xrange(N):
        c = chromosome(n_gene, n_gene*MAXLINKS)
        while not c.Valid():
#            print i, "Not Valid!"
            c = chromosome(n_gene, n_gene*MAXLINKS)
        Chroms.append(c)

def Sort(Chroms, N):
    temp=[]
    for i in range(N):
        temp.append((Chroms[i].fitness, i))
    temp.sort()
    temp.reverse()
    temp1 = []
    for t in temp:  # append the sorted Chromosomes
        temp1.append(Chroms[t[1]])
    return temp1
    
def Cross_Over(Chroms, N, Average_Fitness):
    n_gene=Chroms[0].n_gene
    g = random.Random()
    N_Cross = int( N * CrossOverRate )
    cnt=0
    ChromPool = range(N, 2*N)
    while cnt < N_Cross:
#    for i in range(N-30):
#        if cnt>=N_Cross:
#            break
        # select two chromosomes
        i = g.randint(0, len(ChromPool)-1)
        c1 = ChromPool[i];  del ChromPool[i]
        j = g.randint(0, len(ChromPool)-1)
        c2 = ChromPool[j];  del ChromPool[j]
#        p = (Chroms[c1].fitness / Average_Fitness) * 0.6
#        if g.random() > p:
#            continue
        p = g.randint(1, Chroms[c1].n_cells-1)  # One point CrossOver -- avoid end points
        C1 = chromosome(n_gene, n_gene*n_gene, Chroms[c1].cells[:p]+Chroms[c2].cells[p:], Chroms[c1].Wij[:p]+Chroms[c2].Wij[p:], Chroms[c1].alpha[:], Chroms[c1].beta[:])
        C2 = chromosome(n_gene, n_gene*n_gene, Chroms[c2].cells[:p]+Chroms[c1].cells[p:], Chroms[c2].Wij[:p]+Chroms[c1].Wij[p:], Chroms[c2].alpha[:], Chroms[c2].beta[:])
#        C1.ChecknMaintainInDegree()
#        C2.ChecknMaintainInDegree()
        Chroms[c1] = C1
        while not Chroms[c1].Valid():
#            print i, "Not Valid!"
            Chroms[c1] = chromosome(n_gene, n_gene*n_gene)
        Chroms[c2] = C2
        while not Chroms[c2].Valid():
#            print i, "Not Valid!"
            Chroms[c2] = chromosome(n_gene, n_gene*n_gene)
        cnt+=1

def Mutation(Chroms, N, Average_Fitness):
    n_gene = Chroms[0].n_gene
    g = random.Random()
    for i in range(N, 2*N):
        if Chroms[i].fitness:   # ignore Chromosome with fitness==0
            p = ( Chroms[i].fitness / Average_Fitness ) * 0.5
            if g.random() < p:
                Chroms[i].Mutate()
            while not Chroms[i].Valid():
    #            print i, "Not Valid!"
                Chroms[i] = chromosome(n_gene, n_gene*n_gene)


def Intermediate(Chroms, N, Ave_fit):    # generate intermediate generation
    n_gene = Chroms[0].n_gene
    g = random.Random()
    Relative_Fitness=[]
    temp=[]
    for i in xrange(N):
        Relative_Fitness.append( (Ave_fit/(Chroms[i].fitness - 0.1)) )
        print 'chromosome', i, Chroms[i].fitness, Ave_fit, Relative_Fitness[i]
        temp.append((Relative_Fitness[i], i))
    temp.sort()
    temp.reverse()
    for t in temp:  # process chromosome with higher fitness
        i = t[1]
        n_copy=int(Relative_Fitness[i])
        Relative_Fitness[i]-=n_copy
        if n_copy:  # n_copy != 0
            for j in range(n_copy): # duplicating
                if len(Chroms)==2*N:   # Intermediate chromosomes ready
                    break
                Chroms.append(chromosome(n_gene, n_gene*n_gene, Chroms[i].cells[:], Chroms[i].Wij[:], Chroms[i].alpha[:], Chroms[i].beta[:], Chroms[i].fitness, Chroms[i].AIC, Chroms[i].BIC, Chroms[i].SSE, Chroms[i].RSSE, Chroms[i].SSEVAR))
    for i in xrange(N):
        if len(Chroms)==2*N:   # Intermediate chromosomes ready
            break
        if g.random() > Relative_Fitness[i]:
            continue
        Chroms.append(chromosome(n_gene, n_gene*n_gene, Chroms[i].cells[:], Chroms[i].Wij[:], Chroms[i].alpha[:], Chroms[i].beta[:], Chroms[i].fitness, Chroms[i].AIC, Chroms[i].BIC, Chroms[i].SSE, Chroms[i].RSSE, Chroms[i].SSEVAR))
#    del Chroms[:N]  # delete the previous generation

#   keep the previous generation
    for i in xrange(2*N-len(Chroms)):   # Not enough
        Chroms.append(chromosome(n_gene, n_gene*n_gene))
        
    for i in xrange(N, 2*N):
        while not Chroms[i].Valid():
#            print i, "Not Valid!"
            Chroms[i] = chromosome(n_gene, n_gene*n_gene)
            

def Next_Generation(Chroms, N):
    sum = 0
    ChromsP = Sort(Chroms[:N], N)   # sorted Parent Chromosomes
    ChromsC = Sort(Chroms[N:], N)   # sorted Children
    N_Parent = int(P_Parent * N)
    N_Children = N - N_Parent
    for i in xrange(N_Parent):
        Chroms.append(ChromsP[i])
        sum += ChromsP[i].fitness
    for i in xrange(N_Children):
        Chroms.append(ChromsC[i])
        sum += ChromsP[i].fitness
    del Chroms[:2*N]
    del ChromsP
    del ChromsC
    return sum
    

class WorkerStopToken:  # used to notify the worker to stop
        pass


class SSHWorker(Thread):
    def __init__(self, N_Gene, job_queue, result_queue, cpu_time_queue, host, username, password):
        Thread.__init__(self)
        self.ngene = N_Gene
        self.job_queue = job_queue
        self.result_queue = result_queue
        self.cpu_time_queue = cpu_time_queue
        self.host = host
        self.username = username
        self.password = password
        # s = self.host + str(random.randint(1,100))
        s = self.host
        self.inputfilename = s + '.in'
        self.outputfilename = s + '.out'

    def run(self):
        global programn
        self.connect()  # connect to worker
        self.reconnect = 0    
        # launch the program to compute fitness
#        programn_old = programn
#        self.ssh.sendline('./%s %s' % (programn, trainingfilen))
        self.IOLock = thread.allocate_lock()
        self.Lock = thread.allocate_lock()
        self.jobcnt = 0
        terminate = 0

        id = ''
        
        while not terminate:
            # (i, gene_id, Indeg, combination)
            out = open(self.inputfilename, 'w')
            k = 0
            for cnt in xrange(10):
                try:
                    (id, gene_id, n_links, Links) = self.job_queue.get_nowait()  # block if the queue is empty
                except Queue.Empty:
                    break
                if id is WorkerStopToken:   # notify the worker to quit
                    terminate = 1
                    self.job_queue.put((id, 0, 0, []))   # put the stop token back to job queue
                    break
#                n_links += 1    # add self link
#                Links.append(gene_id)   # self link
                k += 1
                out.write('%s\n' % (id,))
                out.write('%s\n' % (gene_id,))
                out.write('%s\n' % (n_links,))
                for link in Links:
                    out.write('%s\n' % (link,))
            out.write('QUIT\n')
            out.close()
            if not k:   # nothing to do
                if id is WorkerStopToken:
                    break
                else:
                    time.sleep(10)
                    continue
#            if self.reconnect:
#                print 'Try to reconnect', self.host
#                self.reconnect = 0
#                self.ssh.close()
#                self.connect()
#                self.ssh.sendline('./%s %s' % (programn, trainingfilen))
#            if programn != programn_old:
#                programn_old = programn
#                self.ssh.sendline('QUIT')    # notify the program for fitness calculation
#                self.ssh.sendline('./%s %s' % (programn, trainingfilen))
            self.run_one()
            time.sleep(3)
        self.ssh.sendline("exit")

    def connect(self):
        self.ssh = ssh = pexpect.spawn('ssh -l %s node00' % (self.username))
        index = ssh.expect(['%s@.+:.+>' % (self.username,), '%s@.+:.+$' % (self.username,), 'password:', pexpect.TIMEOUT])
        if index == 3:    # time out
            if ssh.before.find('yes/no') != -1:
                ssh.sendline('yes')
            index = ssh.expect(['password:', pexpect.TIMEOUT])
            if index == 0:
                ssh.sendline(self.password)
                index = ssh.expect(['%s@.+:.+>' % (self.username,), '%s@.+:.+$' % (self.username,), pexpect.TIMEOUT])
            else:
                index = 2                
        elif index == 2:    # enter password
            ssh.sendline(self.password)
            index = ssh.expect(['%s@.+:.+>' % (self.username,), '%s@.+:.+$' % (self.username,), pexpect.TIMEOUT])
        if index == 2:
            print 'login failed', self.host
            return
        else:
            print 'login ok', self.host
            
        ssh.sendline('cd '+os.getcwd())
        index = ssh.expect(['%s@.+:.+>' % (self.username,), '%s@.+:.+$' % (self.username,), pexpect.TIMEOUT])
        
    def run_one(self):  # sequential no need for synchronization
#        print self.host, id, gene_id, n_links, Links
#        self.ssh.sendline('%d' % (id,))   # write chromosome id
#        self.ssh.sendline('%d' % (gene_id,))   # write gene id
#        self.ssh.sendline('%d' % (n_links,))    # write # of links
#        for i in Links:   # write incoming links
#            self.ssh.sendline('%d' % (i,))
#        for i in xrange(10):
#            self.ssh.sendline('')
        print 'processing', self.inputfilename
        os.system('rm -f %s' % (self.outputfilename,))  # remove self.outputfilename if any
        self.ssh.sendline('./%s %s < %s > %s' % (programn, trainingfilen, self.inputfilename, self.outputfilename))
        self.get_result()
        
    def get_result(self):
#        self.Lock.acquire()
#        self.jobcnt+=1
#        self.Lock.release()
#        self.IOLock.acquire() # Lock
#        # Get chromosome id, RSSE, SSE, RSSEVAR

        # wait until program terminated successfully
        index = 1
        while index:
            index = self.ssh.expect([r'Program Successfully Terminated', pexpect.TIMEOUT])
            time.sleep(10)

        fail = 1
        while fail:
            try:
                inp = open(self.outputfilename, 'r')
                fail = 0
            except:
                fail = 1
                time.sleep(1)
        line = inp.readline()
        while line:
            if line.find('CPU TIME:') == 0:
                break;
            temp = map(string.atof, line.strip().split()[1:])
            id = int(temp[0]); RSSE = temp[1]
            SSE = 0
            if len(temp) > 2:
                SSE = temp[2]
            RSSEVAR = 0
            if len(temp) > 3:
                RSSEVAR = temp[3]
            gene_id = string.atoi(inp.readline().strip())
            n_links = string.atoi(inp.readline().strip())
            Links = []
            for i in xrange(n_links):
                Links.append(string.atoi(inp.readline().strip()))
            Wij = []
            for i in xrange(n_links):
                Wij.append(string.atof(inp.readline().strip()))
            alpha = string.atof(inp.readline().strip())
            beta = string.atof(inp.readline().strip())
            print self.host, id, gene_id, SSE, RSSE, n_links, Links
            self.result_queue.put( (self.host, id, gene_id, RSSE, SSE, RSSEVAR, Links, Wij, alpha, beta) )   # put to result to result queue
            line = inp.readline()
        inp.close()
        (clock_ticks, cpu_time) = map(string.atof, line.strip().split('\t')[1:])
        self.cpu_time_queue.put( (clock_ticks, cpu_time) )
        
#        print self.host, id, gene_id, SSE, RSSE, n_links, Links
#        self.result_queue.put( (self.host, id, gene_id, RSSE, SSE, RSSEVAR, Links, Wij, alpha, beta) )   # put to result to result queue
#        self.IOLock.release() # Release                
#        self.Lock.acquire()
#        self.jobcnt-=1
#        self.Lock.release()
#        time.sleep(2)


def get_result(path_output_filename):
    # wait until program terminated successfully
    #while not os.path.exists(self.errfilename):
    #time.sleep(10)

    #fail = 1
    #while fail:
    #    inp = open(self.errfilename, 'r')
    #    lines = inp.readlines()
    #    inp.close()
    #    if not lines:   # lines empty? ==> not successfully terminated?
    #        self.rerun = 1
    #        return
    #    elif lines[-1].find('Program Successfully Terminated') == 0:
    #        fail = 0

    ##### Here is for checking whether "CPU TIME:" exists
    #while not os.path.exists(self.outputfilename):
    #    time.sleep(2)
    #        
    #time.sleep(10)
    #fail = 1
    #while fail:
    #    inp = open(self.outputfilename, 'r')
    #    lines = inp.readlines()
    #    inp.close()
    #    if lines and lines[-1].find('CPU TIME:') == 0:    # found
    #        fail = 0
    #    else:
    #        fail += 1
    #    time.sleep(5)
    #    if fail > 15:
    #        self.rerun = 1
    #        return
    ##### Here is for checking whether "CPU TIME:" exists

    id = -10
    gene_id = -10
    RSSE = -10
    SSE = -10
    RSSEVAR = -10
    Links = []
    Wij = []
    alpha = -10
    beta = -10
    
    inp = open(path_output_filename, 'r')
    line = inp.readline()
    while line:
        if line.find('CPU TIME:') == 0:
            break;
        
        temp = map(string.atof, line.strip().split()[1:])
        
        id = int(temp[0]); RSSE = temp[1]
        SSE = 0
        if len(temp) > 2:
            SSE = temp[2]
        RSSEVAR = 0
        if len(temp) > 3:
            RSSEVAR = temp[3]
        gene_id = string.atoi(inp.readline().strip())
        n_links = string.atoi(inp.readline().strip())
        Links = []
        for i in xrange(n_links):
            Links.append(string.atoi(inp.readline().strip()))
        Wij = []
        for i in xrange(n_links):
            Wij.append(string.atof(inp.readline().strip()))
        alpha = string.atof(inp.readline().strip())
        beta = string.atof(inp.readline().strip())
        #print self.host, id, gene_id, SSE, RSSE, n_links, Links
    #    self.result_queue.put( (self.host, id, gene_id, RSSE, SSE, RSSEVAR, Links, Wij, alpha, beta) )   # put to result to result queue
        line = inp.readline()
    inp.close()
    (clock_ticks, cpu_time) = map(string.atof, line.strip().split('\t')[1:])
    return [(id, gene_id, RSSE, SSE, RSSEVAR, Links, Wij, alpha, beta), (clock_ticks, cpu_time)]
    #self.cpu_time_queue.put( (clock_ticks, cpu_time) )

    
def generate_input_for_SA(generation, chrom_id, gene_id, n_links, Links, input_filename_for_the_job):
    path_input_filename_for_the_job = 'Gen_' + str(generation) + '/' + input_filename_for_the_job
    if os.path.exists(path_input_filename_for_the_job):
        print '%s already existed' % (path_input_filename_for_the_job)
        sys.exit()
    else:
        out = open(path_input_filename_for_the_job, 'w')
        out.write('%s\n' % (chrom_id,))
        out.write('%s\n' % (gene_id,))
        out.write('%s\n' % (n_links,))
        for link in Links:
            out.write('%s\n' % (link,))
        out.write('QUIT\n')
    out.close()
    ## put a checking point here


def decide_script_name(generation, chrom_id, gene_id, Indeg):
    shared_prefix = 'G' + str(generation) + 'C' + str(chrom_id) + 'g' + str(gene_id)
    script_name = shared_prefix + 'Id' + str(Indeg) + '.sh'
    if os.path.exists(script_name):
        print 'used script name: %s' % (script_name)
        sys.exit()
    else:
        return shared_prefix, script_name


## for Scheme Three: all jobs have the same amount of SAs to run.
def collect_SAs_for_one_gene(Generation, chrom_id, gene_id, Indeg, combinations, input_filenames_list, output_filenames_list, collected_list):
	shared_prefix, script_name = decide_script_name(Generation, chrom_id, gene_id, Indeg)
	for i in range(0, len(combinations)):
		# no self-link first
                extend_prefix = shared_prefix
		for j in combinations[i]:
        		extend_prefix = extend_prefix + '_' + str(j)
                input_filename = extend_prefix + '.in'
                output_filename = extend_prefix + '.out'
                output_filenames_list.append( output_filename )
                input_filenames_list.append(input_filename)
                collected_list.append(0)
		generate_input_for_SA(Generation, chrom_id, gene_id, Indeg, combinations[i], input_filename)

        	# with self-link then
	        extend_prefix = shared_prefix
        	for j in combinations[i]:
	        	extend_prefix = extend_prefix + '_' + str(j)
        	extend_prefix = extend_prefix + '_' + str(gene_id)
	        input_filename = extend_prefix + '.in'
	        output_filename = extend_prefix + '.out'
	        output_filenames_list.append( output_filename )
		input_filenames_list.append(input_filename)
	        collected_list.append(0)
	        generate_input_for_SA(Generation, chrom_id, gene_id, Indeg+1, combinations[i][:] + [gene_id], input_filename)


def submit_balance_jobs_for_one_chromosome(generation, chrom_id, input_filenames_list, output_filenames_list):

	shared_prefix = 'G' + str(generation) + 'C' + str(chrom_id)

	total_SA_nodes_number = 10

	if len(input_filenames_list) != len(output_filenames_list):
		print 'len(input_filenames_list) != len(output_filenames_list). Weird.'
		sys.exit()
	if len(input_filenames_list) % total_SA_nodes_number == 0:
		num_per_part = len(input_filenames_list)/total_SA_nodes_number
	else:
		num_per_part = len(input_filenames_list)/total_SA_nodes_number + 1

        job_partitions_input = []
        job_partitions_output = []
        for i in range(0, total_SA_nodes_number):
            job_partitions_input.append(input_filenames_list[i*num_per_part:(i+1)*num_per_part])
            job_partitions_output.append(output_filenames_list[i*num_per_part:(i+1)*num_per_part])
                    

	for i in range(0, total_SA_nodes_number):
		script_name = shared_prefix + 'B' + 'Set' + str(i) + '.sh'
		if os.path.exists(script_name):
			print 'used script name: %s' % (script_name)
			sys.exit()
		out_script = open(script_name, 'w')
		out_script.write('#PBS -N %s\n' % (script_name[:-3]) )
		out_script.write('#PBS -e %s.e\n' % (script_name[:-3]))
		out_script.write('#PBS -o %s.o\n' % (script_name[:-3]))
		out_script.write('cd %s/Gen_%d\n' % (main_path, generation)) ## Modify this part to set up proper dirs.

		for j in range(0, len(job_partitions_input[i])):
			if job_partitions_input[i][j][0:-3] != job_partitions_output[i][j][0:-4]:
				print 'ERR: input name %s is different from output name %s.' % (job_partitions_input[i][j][0:-3], job_partitions_output[i][j][0:-4])
				sys.exit()

			out_script.write('./%s %s < %s > %s\n' % (programn, trainingfilen, job_partitions_input[i][j], job_partitions_output[i][j]))
		out_script.close()

		## login to main node, then submit jobs ##
		#os.system('qsub -l nodes=1 %s' % (script_name))
                ssh = pexpect.spawn('ssh -l %s node00' % (username))
                index = ssh.expect(['%s@.+:.+>' % (username,), '%s@.+:.+$' % (username,), 'password:', pexpect.TIMEOUT])
                if index == 3:    # time out
                    if ssh.before.find('yes/no') != -1:
                        ssh.sendline('yes')
                        index = ssh.expect(['password:', pexpect.TIMEOUT])
                        if index == 0:
                            ssh.sendline(password)
                            index = ssh.expect(['%s@.+:.+>' % (username,), '%s@.+:.+$' % (username,), pexpect.TIMEOUT])
                        else:
                            index = 2
                elif index == 2:    # enter password
                    ssh.sendline(password)
                    index = ssh.expect(['%s@.+:.+>' % (username,), '%s@.+:.+$' % (username,), pexpect.TIMEOUT])
                    if index == 2:
                        print 'login failed'
                else:
                    print 'login ok'
                ssh.sendline('cd %s' % (main_path))
                index = ssh.expect(['%s@.+:.+>' % (username,), '%s@.+:.+$' % (username,), pexpect.TIMEOUT])
                if index == 1 or index == 2:
                    ssh.sendline('qsub -l nodes=1 %s' % (script_name))
                    index = ssh.expect(['%s@.+:.+>' % (username,), '%s@.+:.+$' % (username,), pexpect.TIMEOUT])
                    if index == 3:
                        print 'time out'
                else:
                    print 'time out'


def collect_results_for_one_chromosome(output_filename_list, collected_list, results_list, cpu_time_list, start_time, generation, chrom_id):
	wait_counters = [0]*len(output_filename_list)
	grep_counter = 0
	collected_cnt = len(output_filename_list)
	while collected_cnt:
		for k in range(0, len(output_filename_list)):
			if collected_list[k] == 0: # not yet collected
				path_output_filename = 'Gen_' + str(Generation) + '/' + output_filename_list[k]
				if os.path.exists(path_output_filename):
					tmp_command = 'grep "CPU TIME" ' + path_output_filename + ' > mid.tmp'
					grep_counter = grep_counter + 1
					x = os.system(tmp_command)
					if x == 0: # already finished the SA computation, so collect it!
						collected_list[k] = 1
						collected_cnt = collected_cnt - 1
						results_N_times = get_result(path_output_filename)
						results_list.append( results_N_times[0] )
						cpu_time_list.append( results_N_times[1] )

					else:
                                                current_time = time.time()
                                                if current_time - start_time > 20*60: ## longer than 50 mins., consider to rerun
                                                    os.system('ls G%dC%dBSet*.e > finished_jobs.txt' % (generation, chrom_id))
                                                    f_fin = open('finished_jobs.txt', 'r')
                                                    fin_lines = f_fin.readlines()
                                                    f_fin.close()
                                                    if len(fin_lines) == 10: ## all jobs are finished
                                                        ## rerun the entry
                                                        os.system('rm %s' % (path_output_filename))
                                                        os.system('./%s %s < %s > %s' % (programn, trainingfilen, path_output_filename[0:-4]+'.in', path_output_filename))
                                                        print 'rerun for %s' % (path_output_filename)
                                                        time.sleep(15)
                                                        tmp_command = 'grep "CPU TIME" ' + path_output_filename + ' > mid.tmp'
                                                        grep_counter = grep_counter + 1
                                                        x = os.system(tmp_command)
                                                        while x != 0: # not yet finished the SA computation.
                                                            time.sleep(5)
                                                            x = os.system(tmp_command)
                                                    else:
                                                        wait_counters[k] = wait_counters[k] + 1
                                                        time.sleep(5)
                                                else:
                                                    wait_counters[k] = wait_counters[k] + 1
                                                    time.sleep(5)
				else:
					wait_counters[k] = wait_counters[k] + 1
					time.sleep(10)
				
				if wait_counters[k] == 10000:
					print '%s has been collected for 10000 times.' % (output_filename_list[k])
			time.sleep(1)
		if grep_counter % 10000 == 0:
			print 'grep_counter: %d' % (grep_counter)

## NOTE: the output formular for Structural Penalty was modified 
##       to be 2*math.log(N_TimePoint)*((N_Gene - N_FACTOR)*2 
def Write_Log(Chroms, Generation, N_Gene, N, idx_max, out, outSSE, outSSEALL, Gclock_ticks, Gcpu_time):
    outSSE.write('%s\t%s\n' % (Chroms[idx_max].fitness, Chroms[idx_max].BIC))
    out.write('Generation:\t%s\tClock Ticks:\t%s\tCPU Time (sec):\t%s\n' % (Generation, Gclock_ticks, Gcpu_time))
    # structural penalty in AIC
    out.write('SSEVAR&Structural Penalty&RSSE&SSE&BIC&AIC:\t%s\t%s\t%s\t%s\t%s\t%s\n' \
              % (Chroms[idx_max].SSEVAR, 2*math.log(N_TimePoint)*((N_Gene - N_FACTOR)*2 + Chroms[idx_max].narcs ), \
                 Chroms[idx_max].RSSE, Chroms[idx_max].SSE, Chroms[idx_max].BIC, Chroms[idx_max].AIC))
    out.write('Wij:\n')
    for i in range(N_Gene):
        for j in range(N_Gene):
            if j:
                out.write('\t')
            out.write('%s' % (Chroms[idx_max].Wij[i*N_Gene+j],))
        out.write('\n')
    out.write('Alpha:\n')   # alpha
    out.write('\t'.join(map(str, Chroms[idx_max].alpha))+'\n')
    out.write('Beta:\n')    # beta
    out.write('\t'.join(map(str, Chroms[idx_max].beta))+'\n')
    # Write the SSE of one generation
    tmp=[]
    for i in range(N):  # for each chromosome
        tmp.append('%s\t%s\t%s' % (Chroms[i].fitness, Chroms[i].BIC, Chroms[i].narcs))
    tmp = string.join(tmp, '\t')
    outSSEALL.write(tmp+'\n')
    
    outSSE.flush()
    out.flush()
    outSSEALL.flush()


def argmax(Chroms, N):
    idx_max=0
    for i in range(1, N):
        if Chroms[i].fitness > Chroms[idx_max].fitness:
            idx_max = i
    return idx_max


def SaveCurrentStatus(filen, Chroms, N, Generation, Max):
    out=open(filen + '%s.txt' % (Generation), 'w')
    
    out.write('%s\n' % (Generation,))
    out.write('%s\n' % (Max,))
    if Generation > 1:
        N *= 2
    for i in range(N):
        out.write('%s\n' % (Chroms[i].fitness,))    # fitness -- BIC score or AIC score
        out.write('%s\n' % (Chroms[i].AIC,))    # AIC score
        out.write('%s\n' % (Chroms[i].BIC,))    # BIC score
        out.write('\t'.join(map(str, Chroms[i].cells))+'\n')    # network structure
        out.write('\t'.join(map(str, Chroms[i].Wij))+'\n')  # Wij
        out.write('\t'.join(map(str, Chroms[i].alpha))+'\n')    # alpha
        out.write('\t'.join(map(str, Chroms[i].beta))+'\n')    # beta
        out.flush()
    out.close()        

def ReadStatus(filen, Chroms, N, N_Gene):
    inp=open(filen, 'r')
    
    Generation=string.atoi(inp.readline()[:-1]) # Generation
    Max=string.atof(inp.readline()[:-1])    # Max fitness
    if Generation > 1:
        N *= 2
    for i in range(N):
        fitness=string.atof(inp.readline()[:-1])    # fitness -- BIC score or AIC score
        AIC=string.atof(inp.readline()[:-1])    # AIC score
        BIC=string.atof(inp.readline()[:-1])    # BIC score
        net=map(string.atoi, string.split(inp.readline()[:-1], '\t'))   # network structure
        Wij=map(string.atof, string.split(inp.readline()[:-1], '\t'))   # Wij
        alpha=map(string.atof, string.split(inp.readline()[:-1], '\t')) # alpha
        beta=map(string.atof, string.split(inp.readline()[:-1], '\t'))  # beta
        Chroms.append(chromosome(N_Gene, N_Gene*N_Gene, net, Wij, alpha, beta, fitness, AIC, BIC))
    inp.close()
    return (Generation, Max)


def Combinations(gene_id, N_Gene, n_links):
    temp = [[]]
    for i in xrange(n_links):
        l = len(temp)
        for j in xrange(l):
            if len(temp[j]):
                begin = temp[j][-1] + 1
            else:
                begin = 0
            for k in xrange(begin, N_Gene):
                if k == gene_id:
                    continue
                temp.append(temp[j] + [k])
        del temp[:l]
    return temp

    
# Initialize job queue and result queue
#job_queue = Queue.Queue(0)
#result_queue = Queue.Queue(0)
#cpu_time_queue = Queue.Queue(0)

# read microarray data
MicroArray=[]
(N_TimePoint, N_Gene)=My_Sub.Read_MicroArray(argv[1], MicroArray)
print N_TimePoint, N_Gene

RSSE_GENE = {}
for i in xrange(N_FACTOR, N_Gene):
    RSSE_GENE[i] = {}

# Initialize telnet workers
# username = getpass.getuser()
#password = getpass.getpass()
username = 'XXXXX'
password = 'XXXXX'
#for host in ssh_workers:
#    SSHWorker(N_Gene, job_queue, result_queue, cpu_time_queue, host, username, password).start()

# Main function starts here
N_Chrom = 10
Chroms = []
StatusFilen = 'Status_Generation'
# Initialization
Max_old=-1e11
if len(argv)==2:    # normal initialization
    Init_Chroms(Chroms, N_Chrom, N_Gene)
    Max=-1e10
    Generation=0
    #open Log file for write
    out=open(argv[0]+'.Log', 'w')
    outSSE=open(argv[0]+'.SSELog', 'w')
    outSSEALL=open(argv[0]+'.SSEALLLog', 'w')    
#elif len(argv)==3:  # continuing process
#    (Generation, Max)=ReadStatus(argv[2], Chroms, N_Chrom, N_Gene)
#    #open Log file for write
#    out=open(argv[0]+'.Log', 'a')
#    outSSE=open(argv[0]+'.SSELog', 'a')
#    outSSEALL=open(argv[0]+'.SSEALLLog', 'a')

elif len(argv)==4:  # continuing process
    (Generation, Max)=ReadStatus(argv[2], Chroms, N_Chrom, N_Gene)
    #open Log file for write
    out=open(argv[0]+'.Log', 'a')
    outSSE=open(argv[0]+'.SSELog', 'a')
    outSSEALL=open(argv[0]+'.SSEALLLog', 'a')

    # using cPickle to load RSSE here.
    # remember to modify the argv condition too.
    inRSSE_GENE=open(argv[3], 'r')
    RSSE_GENE = cPickle.load(inRSSE_GENE)
    inRSSE_GENE.close()

else:
    print 'Usage: python %s <microarray data filename> <latest Status_Generation filename (opt.)> <latest RSSE_GENE filename (opt.)>' % (argv[0])
    sys.exit()


# Check for the existence of .rec pickle file
filen = argv[0]+'.RSSE_GENE.rec'
if os.path.exists(filen):
    inRSSE_GENE=open(filen, 'r')
    RSSE_GENE = cPickle.load(inRSSE_GENE)
    inRSSE_GENE.close()


#last_total_Job_Cnt = 0
while Max > Max_old or Generation < 10000:
    Gclock_ticks = 0
    Gcpu_time = 0
    Generation+=1

    #if Generation > 1:
    #    ## remove a.out and the MArray data in the previous generation folder.
    #    os.system('rm %s/Gen_%d/%s' % (main_path, Generation-1, programn))
    #    os.system('rm %s/Gen_%d/%s' % (main_path, Generation-1, trainingfilen))
    #    ## zip all .in/.out files in the previous generation folder.
    #    if last_total_Job_Cnt > 0:
    #        os.system('gzip %s/Gen_%d/*' % (main_path, Generation-1))

    ## de thread
    if os.path.exists('Gen_%d' % (Generation)):
        print 'the dir. Gen_%d already exists' % (Generation)
        sys.exit()
    else:
        os.system('mkdir Gen_%d' % (Generation))
        os.system('cp %s Gen_%d' % (programn, Generation))
        os.system('cp %s Gen_%d' % (argv[1], Generation))
    ## de thread

    if Generation > 100:
        programn = program_slow
    Max_old=Max
    # Computing SSE + ln(1+narcs)
    print 'Start Computing the SSE for each chromosome'
    if Generation == 1:
        chrom_ids = xrange(N_Chrom)
    else:   # Generation 2 and above
        chrom_ids = xrange(N_Chrom, 2*N_Chrom)
    Average_Fitness=0.0

    # YC for de thread
    cpu_time_list = []
    total_Job_Cnt = 0
    # YC for de thread

    for i in chrom_ids:   # for each chromosome
        Job_Cnt = 0

        # YC for de thread
        collected_list = []
        results_list = []
        output_filename_list = []
	input_filename_list = []
        # YC for de thread

        print 'Processing chromosome%s...' % (i,)
        for gene_id in xrange(N_FACTOR, N_Gene):    # for each gene
            Indeg = Chroms[i].InDeg[gene_id]
            if not RSSE_GENE[gene_id].has_key(Indeg):   # not computed yet
                # compute each combination
                combinations = Combinations(gene_id, N_Gene, Indeg)
		Job_Cnt = Job_Cnt + len(combinations)*2
		total_Job_Cnt = total_Job_Cnt + len(combinations)*2
		collect_SAs_for_one_gene(Generation, i, gene_id, Indeg, combinations, input_filename_list, output_filename_list, collected_list)
            else:
                print 'already computed %d %d' % (gene_id, Indeg)

            print 'Gene: %d, Job_Cnt: %d' % (gene_id, Job_Cnt)

        # YC for de thread
        # start to collect results
	if len(input_filename_list) != 0:
		time.sleep(10)
		submit_balance_jobs_for_one_chromosome(Generation, i, input_filename_list, output_filename_list)
		time.sleep(200)
                start_time = time.time()
                collect_results_for_one_chromosome(output_filename_list, collected_list, results_list, cpu_time_list, start_time, Generation, i)

        for k in range(0, len(results_list)):
            (id, gene_id, RSSE, SSE, RSSEVAR, Links, Wij, alpha, beta) = results_list[k]
            Indeg = len(Links)
            if gene_id in Links:    # contains self-link
                Indeg -= 1
            if not RSSE_GENE[gene_id].has_key(Indeg):
                RSSE_GENE[gene_id][Indeg] = []
            RSSE_GENE[gene_id][Indeg].append((RSSEVAR, SSE, RSSE, Links, Wij, alpha, beta)) # sort wrt SSEVAR
            
            ## Everytime RSSE_GENE is updated, open an file to store it.
            outRSSE_GENE = open(argv[0]+'.RSSE_GENE.rec', 'w')
            cPickle.dump(RSSE_GENE, outRSSE_GENE)
            outRSSE_GENE.close()
            ## Everytime RSSE_GENE is updated, open an file to store it.

#                for combination in combinations:
#                    Job_Cnt += 1
#                    job_queue.put((i, gene_id, Indeg, combination))
#                    Job_Cnt += 1
#                    job_queue.put((i, gene_id, Indeg + 1, combination[:] + [gene_id]))
#        while Job_Cnt:  # (self.host, id, gene_id, RSSE, SSE, RSSEVAR, Links, Wij, alpha, beta)
#            (host, id, gene_id, RSSE, SSE, RSSEVAR, Links, Wij, alpha, beta) = result_queue.get()
#            print host, id, gene_id, SSE, RSSE, RSSEVAR, Job_Cnt, 'to go...'
#            Job_Cnt -= 1
#            Indeg = len(Links)
#            if gene_id in Links:    # contains self-link
#                Indeg -= 1
#            if not RSSE_GENE[gene_id].has_key(Indeg):
#                RSSE_GENE[gene_id][Indeg] = []
#            RSSE_GENE[gene_id][Indeg].append((RSSEVAR, SSE, RSSE, Links, Wij, alpha, beta)) # sort wrt SSEVAR
        print 'Finished computing chromosome%s...' % (i,)
#        while 1:
#            try:
#                (clock_ticks, cpu_time) = cpu_time_queue.get_nowait()  # block if the queue is empty
#                Gclock_ticks += clock_ticks
#                Gcpu_time += cpu_time
#            except Queue.Empty:
#                break
        
        for k in cpu_time_list:
            Gclock_ticks = Gclock_ticks + k[0]
            Gcpu_time = Gcpu_time + k[1]

        # Reset parameters
        Chroms[i].Wij = [0.0 for k in xrange(N_Gene*N_Gene)]    # i influence j
        Chroms[i].alpha = [0.0 for k in xrange(N_Gene)]
        Chroms[i].beta = [0.0 for k in xrange(N_Gene)]
        Chroms[i].SSE = 0.0
        Chroms[i].RSSE = 0.0
        Chroms[i].SSEVAR = 0.0
        for gene_id in xrange(N_FACTOR, N_Gene):    # for each gene
            Indeg = Chroms[i].InDeg[gene_id]
            if len(RSSE_GENE[gene_id][Indeg]) > 1:
                RSSE_GENE[gene_id][Indeg].sort()
                del RSSE_GENE[gene_id][Indeg][1:]   # retain only one combination
            (RSSEVAR, SSE, RSSE, Links, Wij, alpha, beta) = RSSE_GENE[gene_id][Indeg][0]
            Chroms[i].SSEVAR += RSSEVAR
            Chroms[i].SSE += SSE
            Chroms[i].RSSE += RSSE
            # ignore RSSEVAR for now
            for j in xrange(len(Links)): # Links[j] influence gene_id
                Chroms[i].Wij[Links[j]*N_Gene + gene_id] = Wij[j]
            Chroms[i].alpha[gene_id] = alpha
            Chroms[i].beta[gene_id] = beta
            if gene_id in Links:    # self-link is predicted ==> parameter penalty
                Chroms[i].narcs += 1
        Chroms[i].BIC =  Chroms[i].SSEVAR * W_SSE + 2*math.log(N_TimePoint)*((N_Gene - N_FACTOR)*2 + Chroms[i].narcs )  # SSE / sigma^2 + ln(D)*k, D: number of data points
        Chroms[i].AIC =  Chroms[i].SSEVAR * W_SSE + 2*((N_Gene - N_FACTOR)*2 + Chroms[i].narcs)  # SSE / sigma^2 + 2k, k: total number of unknowns
        Chroms[i].fitness = -Chroms[i].BIC   # use - BIC as fitness
        Average_Fitness += Chroms[i].fitness

    sys.stdout.flush()  # flush STDOUT
    
    if Generation == 1:
        Average_Fitness /= N_Chrom  # Compute Average Fitness
    else:   # Generation 2 and above
        Average_Fitness = Next_Generation(Chroms, N_Chrom) / N_Chrom
    idx_max = argmax(Chroms, N_Chrom)
    Max = Chroms[idx_max].fitness
    
    print 'Maximal Fitness (BIC or AIC):', Max
    ##### Write generation Max to Log file ########
    Write_Log(Chroms, Generation, N_Gene, N_Chrom, idx_max, out, outSSE, outSSEALL, Gclock_ticks, Gcpu_time)
    print 'Generating Intermediate population'
    Intermediate(Chroms, N_Chrom, Average_Fitness)
    print 'Performing Cross Over'    
    Cross_Over(Chroms, N_Chrom, Average_Fitness)
    print 'Performing Mutation'
    Mutation(Chroms, N_Chrom, Average_Fitness)
    # Save Current Status
    if Generation % 10 == 0:
        SaveCurrentStatus(StatusFilen, Chroms, N_Chrom, Generation, Max)

    ## remove a.out and the MArray data in the previous generation folder.
    os.system('rm %s/Gen_%d/%s' % (main_path, Generation, programn))
    os.system('rm %s/Gen_%d/%s' % (main_path, Generation, trainingfilen))

    if total_Job_Cnt > 0:
	os.system('gzip %s/Gen_%d/*' % (main_path, Generation))
	os.system('tar -cvf %s/Gen_%d/Gen%d.tar %s/Gen_%d/*.gz' % (main_path, Generation, Generation, main_path, Generation))
	os.system('rm %s/Gen_%d/*.gz'% (main_path, Generation))
        os.system('gzip %s/G%dC*Set*' % (main_path, Generation))
        os.system('tar -cvf G%d.tar *.gz' % (Generation))
        os.system('rm *.gz')
    else:
        os.system('rmdir %s/Gen_%d' % (main_path, Generation))

#job_queue.put((WorkerStopToken, 0, 0, []))

outl = open(argv[0] + '.FINAL', 'w')
for gene_id in xrange(N_Gene):
    outl.write('Gene%s:\n' % (gene_id,))
    for Indeg in xrange(1, MAXLINKS+1):
        #(AIC, SSE, RSSE, RSSEVAR, Links, Wij, alpha, beta) = RSSE_GENE[gene_id][Indeg][0]
        (RSSEVAR, SSE, RSSE, Links, Wij, alpha, beta) = RSSE_GENE[gene_id][Indeg][0]
        BIC = RSSEVAR*W_SSE + 2*math.log(N_TimePoint)*(2 + len(Links))
        AIC = RSSEVAR*W_SSE + 2*(2 + len(Links))
        
        outl.write('\tIn Degree: %s\n' % (Indeg,))
        #outl.write('\t\tAIC: %s\n' % (AIC,))
        outl.write('\t\tBIC: %s\n' % (BIC,))
        #outl.write('\t\tSSE: %s\n' % (SSE,))
        outl.write('\t\tSSEVAR: %s\n' % (RSSEVAR,))
        outl.write('\t\tLinks: %s\n' % ('\t'.join(map(str, Links)),))
outl.close()
         
out.close()
outSSE.close()
outSSEALL.close()
