######################################################################
## 
## Some subroutines and functions
## 
######################################################################
import string
import random
import math

# input an empty array MicroArray=[]
# Read MicroArray Data from file
# Return (#Time points, #Genes)
def Read_MicroArray(filen, MicroArray):
    inp=open(filen, 'r')
    line=inp.readline()
    cnt = []
    t = 0   # number of time points
    while line:
        if not line.strip():    # skip empty line
            # empty lines separate replicates
            cnt.append(t)
            t = 0
            line = inp.readline()
            continue
        # must be tab-separated
        MicroArray.append(map(string.atof, line.strip().split('\t')))
        t += 1
        line = inp.readline()
    inp.close()
    cnt.append(t)
    # #Time points in case of mutiple replicates
    # When each replicate has the same T ==> return T?
    # When each replicate has different Ti ==> return Sum(Ti)?
    tp = 0
    for t in cnt:
        tp += t
    tp /= len(cnt)
    TimePoint = tp
    for t in cnt:
        if t != tp:
            TimePoint = tp * len(cnt)
    return (TimePoint, len(MicroArray[0]))


def Predict(chrom, MicroArray, Predicted):
    n_time_points=len(MicroArray)
    n_gene=len(MicroArray[0])    
    Predicted.append([])    # at time 0, expression can't be predicted
    for t in range(1, n_time_points):  # for each time points
        Predicted.append([])
        for i in range(n_gene):    # Gi(t) = Sum( W[j][i] gj(t-1) + aj )
            Predicted[t].append(0.0)
            for j in range(n_gene):
                Predicted[t][i]+= ( chrom.Wij[j*n_gene+i] * MicroArray[t-1][j] + chrom.alpha[j] )

def Cal_Energy(chrom, MicroArray):  # SSE
    SSE=0.0
    n_time_points=len(MicroArray)
    n_gene=len(MicroArray[0])
    # predicting gene expressions
    Predicted=[]
    Predict(chrom, MicroArray, Predicted)
    for t in range(1, n_time_points):
        for i in range(n_gene):
            SSE += (Predicted[t][i]-MicroArray[t][i]) * (Predicted[t][i]-MicroArray[t][i])
#    print '%s\t%s' % (SSE, math.log(1+chrom.narcs))
    return SSE + math.log(1 + chrom.narcs)


# Test Read_MicroArray()
#MicroArray=[]
#from sys import argv
#if len(argv) > 1:
#    print 'Testing Read_MicroArray()...'
#    print '(#Time points, #Genes) =', Read_MicroArray(argv[1], MicroArray)
