######################################################################
## 
## Process and Visualize GeneNet output
## command line: GeneNet_XXXX.Log, MicroArray data, MicroArray data(one gene/line),
##               XXXX.names(true network)
######################################################################
import sys
from sys import argv
import string
import os



CPUtime = {}
ClockTicks = {}
SSEs = []
Gen = 0 # Generation
buffer = []
inp = open(argv[1], 'r')
for line in inp.readlines():
    if line.find('Generation:')==0:
        if buffer:  # Process the previous generation
            # Generation:\t1\tClock Ticks:\t2314760000.0\tCPU Time (sec):\t2314.76
            cells = buffer[0].strip().split('\t')
            gen = string.atoi(cells[1])
            ClockTicks[gen] = string.atof(cells[3])
            CPUtime[gen] = string.atof(cells[5])
            # SSEVAR&Structural Penalty&RSSE&SSE&BIC&AIC:	2360.199	44	1.453134	381.2569	2536.91806727	2404.199
            cells = buffer[1].strip().split('\t')
            SSEkeys = cells[0][:-1].split('&')
            SSEs.append(map(string.atof, cells[1:]))
            
        buffer = []
        Gen = string.atoi(line.strip().split('\t')[1])
    buffer.append(line)
inp.close()

# Process the last generation
# Generation:\t1\tClock Ticks:\t2314760000.0\tCPU Time (sec):\t2314.76
cells = buffer[0].strip().split('\t')
gen = string.atoi(cells[1])
ClockTicks[gen] = string.atof(cells[3])
CPUtime[gen] = string.atof(cells[5])
# SSEVAR&Structural Penalty&RSSE&SSE&BIC&AIC:	2360.199	44	1.453134	381.2569	2536.91806727	2404.199
cells = buffer[1].strip().split('\t')
SSEkeys = cells[0][:-1].split('&')
SSEs.append(map(string.atof, cells[1:]))


# produce .genXX file
genfile = argv[1]+'.gen%s' % (Gen,)
out = open(genfile, 'w')
out.writelines(buffer)
out.close()

# produce .SAInpGenXX file
sainpfile = argv[1]+'.SAInpGen%s' % (Gen,)
os.system('python ./Convert2SAnetInp.py %s > %s' % (genfile, sainpfile))

# produce XXXX.para and XXXX.names  -- argv[1]+.genXXX.para, argv[1]+.genXXX.names
os.system('python ./Convert2YCs.py %s %s' % (argv[4], genfile))

# produce the Predicted Expression
predictedfile = genfile+'.PredictedExpression.txt'
os.system('./Predicted_Expression %s %s < %s > %s' % (argv[2], genfile+'.para', sainpfile, predictedfile))

# Read Gene Names
Genes = []
inp = open(argv[4], 'r')
for line in inp.readlines():
    cells = line[:-1].split('\t')
    Genes.append(cells[0])
inp.close()

# Compute TP FP TN FN
TrueNet = {}
Genes = []
inp = open(argv[4], 'r')
for line in inp.readlines():
    cells = line.strip().split('\t')
    Genes.append(cells[0])
    TrueNet[cells[0]] = []
    for cell in cells[1:]:
        TrueNet[cells[0]].append(cell)
inp.close()

print 'TrueNet:'
print TrueNet

PredictedNet = {}
inp = open(genfile+'.names', 'r')
for line in inp.readlines():
    cells = line.strip().split('\t')
    PredictedNet[cells[0]] = []
    for cell in cells[1:]:
        PredictedNet[cells[0]].append(cell)
inp.close()

print 'PredictedNet:'
print PredictedNet

TP = 0
FP = 0
TN = 0
FN = 0
for gi in Genes:
    for gj in Genes:
        if gj in TrueNet[gi] and gj in PredictedNet[gi]:
            TP += 1
        elif not (gj in TrueNet[gi] or gj in PredictedNet[gi]):
            TN += 1
        elif not gj in TrueNet[gi] and gj in PredictedNet[gi]:
            FP += 1
        else:   # gj in TrueNet[gi] and not gj in PredictedNet[gi]:
            FN += 1
Recall = float(TP) / float(TP + FN) # = sensitivity?
Precision = float(TP) / float(TP + FP)
Specificity = float(TN) / float(FP + TN)
Accuracy = float(TP + TN) / float(TP + FN + TN + FP)


# Output
outfile = argv[1]+'.Report%s' % (Gen,)
out = open(outfile, 'w')
# TP FP TN FN Recall Precision Specificity Accuracy
out.write('TP (True Positive)\t\tFP (False Positive)\t\tTN (True Negative)\t\tFN (False Negative)\n')
out.write('%s\t\t%s\t\t%s\t\t%s\n' % (TP, FP, TN, FN))
out.write('Total Positive (= TP + FN): %s\n' % (TP + FN,))
out.write('Total Negative (= TN + FP): %s\n' % (TN + FP,))
out.write('Recall(Sensitivity) = TP / (TP + FN): %s\n' % (Recall,))
out.write('Precision = TP / (TP + FP): %s\n' % (Precision,))
out.write('Specificity = TN / (FP + TN): %s\n' % (Specificity,))
out.write('Accuracy = (TP + TN) / (TP + FN + TN + FP): %s\n' % (Accuracy,))
# CPU time
out.write('Total CPU time: %s seconds\n' % (sum(CPUtime.values()),))

inp = open(genfile, 'r')    # Write genfile
out.write(inp.read())
inp.close()
out.write('True Network:\n')
inp = open(argv[4], 'r')   # Write true net
out.write(inp.read())
inp.close()
out.write('Predicted Network\n')
inp = open(genfile+'.names', 'r')   # Write .names
out.write(inp.read())
inp.close()
inp = open(genfile+'.para', 'r')   # Write .para
out.write(inp.read())
inp.close()

inp = open(argv[3], 'r')    # MicroArray -- one gene per line
inp1 = open(predictedfile, 'r')
for line in inp.readlines():
    line1 = inp1.readline()
    if line == '\n':    # empty line
        continue
    gene = Genes[0]
    Genes.append(gene)  # append gene to the end of the list
    del Genes[0]
    out.write('%s_true\t%s' % (gene, line))
    out.write('%s_predicted\t%s\n' % (gene, line1))
inp.close()
inp1.close()
out.close()

# Another report file
outfile = argv[1]+'.detailedReport%s' % (Gen,)
out = open(outfile, 'w')
gens = CPUtime.keys(); gens.sort()
out.write('Generation\tClock Ticks\tCPU time\t' + '\t'.join(SSEkeys) + '\n')
for i in xrange(len(gens)):
    out.write('%s\t%s\t%s\t%s\n' % \
              (gens[i], ClockTicks[gens[i]], CPUtime[gens[i]], '\t'.join(map(str, SSEs[i]))))
out.close()
